@@ -31,7 +31,8 @@ test_expect_success \
    f4=$(pwd)/4 && d4=$(loop_setup_ "$f4") &&
    vg1=$(this_test_)-1-$$          &&
    vg2=$(this_test_)-2-$$          &&
-   pvcreate $d1 $d2 $d3 $d4'
+   pvcreate $d1 $d2                &&
+   pvcreate --metadatacopies 0 $d3 $d4'
 
 test_expect_success \
   'vgrename normal operation - rename vg1 to vg2' \
@@ -40,6 +41,16 @@ test_expect_success \
    check_vg_field_ $vg2 vg_name $vg2 &&
    vgremove $vg2'
 
+test_expect_success \
+  "vgrename by uuid (bz231187)" '
+   vgcreate $vg1 $d1 $d3 &&
+   UUID=$(vgs --noheading -o vg_uuid $vg1) &&
+   check_vg_field_ $vg1 vg_uuid $UUID &&
+   vgrename $UUID $vg2 &&
+   check_vg_field_ $vg2 vg_name $vg2 &&
+   vgremove $vg2
+'
+
 test_done
 # Local Variables:
 # indent-tabs-mode: nil
